@@ -7,10 +7,10 @@ use crate::compiler::Compiler;
 use crate::error::{Error, ErrorKind};
 use crate::instructions::Instructions;
 use crate::parser::{parse, parse_expr};
-use crate::utils::{AutoEscape, HtmlEscape, RcType};
-use crate::value::{ArgType, FunctionArgs, Value};
+use crate::utils::{AutoEscape, HtmlEscape};
+use crate::value::{ArgType, FunctionArgs, RcType, Value};
 use crate::vm::Vm;
-use crate::{filters, tests};
+use crate::{filters, functions, tests};
 
 /// Represents a handle to a template.
 ///
@@ -140,6 +140,7 @@ pub struct Environment<'source> {
     templates: Source<'source>,
     filters: RcType<BTreeMap<&'source str, filters::BoxedFilter>>,
     tests: RcType<BTreeMap<&'source str, tests::BoxedTest>>,
+    globals: RcType<BTreeMap<&'source str, Value>>,
     default_auto_escape: RcType<dyn Fn(&str) -> AutoEscape + Sync + Send>,
 }
 
@@ -225,6 +226,7 @@ impl<'source> Environment<'source> {
             templates: Source::Borrowed(Default::default()),
             filters: RcType::new(filters::get_builtin_filters()),
             tests: RcType::new(tests::get_builtin_tests()),
+            globals: RcType::new(functions::get_globals()),
             default_auto_escape: RcType::new(default_auto_escape),
         }
     }
@@ -238,6 +240,7 @@ impl<'source> Environment<'source> {
             templates: Source::Borrowed(Default::default()),
             filters: RcType::default(),
             tests: RcType::default(),
+            globals: RcType::default(),
             default_auto_escape: RcType::new(no_auto_escape),
         }
     }
@@ -386,6 +389,34 @@ impl<'source> Environment<'source> {
         RcType::make_mut(&mut self.tests).remove(name);
     }
 
+    /// Adds a new global function.
+    ///
+    /// For details about functions have a look at [`functions`].  Note that
+    /// functions and other global variables share the same namespace.
+    pub fn add_function<F, Rv, Args>(&mut self, name: &'source str, f: F)
+    where
+        Rv: Into<Value>,
+        F: functions::Function<Rv, Args>,
+        Args: FunctionArgs,
+    {
+        self.add_global(name, functions::BoxedFunction::new(f).to_value());
+    }
+
+    /// Adds a global variable.
+    pub fn add_global(&mut self, name: &'source str, value: Value) {
+        RcType::make_mut(&mut self.globals).insert(name, value);
+    }
+
+    /// Removes a global function or variable by name.
+    pub fn remove_global(&mut self, name: &str) {
+        RcType::make_mut(&mut self.globals).remove(name);
+    }
+
+    /// Looks up a function.
+    pub(crate) fn get_global(&self, name: &str) -> Option<Value> {
+        self.globals.get(name).cloned()
+    }
+
     /// Applies a filter with arguments to a value.
     pub(crate) fn apply_filter(
         &self,
@@ -494,3 +525,12 @@ fn test_clone() {
     assert_eq!(env2.get_template("test").unwrap().render(&()).unwrap(), "b");
     assert_eq!(env.get_template("test").unwrap().render(&()).unwrap(), "a");
 }
+
+#[test]
+fn test_globals() {
+    let mut env = Environment::new();
+    env.add_global("a", Value::from(42));
+    env.add_template("test", "{{ a }}").unwrap();
+    let tmpl = env.get_template("test").unwrap();
+    assert_eq!(tmpl.render(()).unwrap(), "42");
+}
