@@ -28,18 +28,17 @@
 //! MiniJinja will perform the necessary conversions automatically via the
 //! [`FunctionArgs`](crate::value::FunctionArgs) and [`Into`] traits.
 use std::collections::BTreeMap;
-use std::sync::Arc;
 
 use crate::environment::Environment;
 use crate::error::Error;
 use crate::utils::HtmlEscape;
-use crate::value::{ArgType, FunctionArgs, Value};
+use crate::value::{ArgType, FunctionArgs, RcType, Value};
 
 type FilterFunc =
     dyn Fn(&Environment, Value, Vec<Value>) -> Result<Value, Error> + Sync + Send + 'static;
 
 #[derive(Clone)]
-pub(crate) struct BoxedFilter(Arc<FilterFunc>);
+pub(crate) struct BoxedFilter(RcType<FilterFunc>);
 
 /// A utility trait that represents filters.
 pub trait Filter<V = Value, Rv = Value, Args = Vec<Value>>: Send + Sync + 'static {
@@ -77,14 +76,16 @@ impl BoxedFilter {
         Rv: Into<Value>,
         Args: FunctionArgs,
     {
-        BoxedFilter(Arc::new(move |env, value, args| -> Result<Value, Error> {
-            f.apply_to(
-                env,
-                ArgType::from_value(Some(value))?,
-                FunctionArgs::from_values(args)?,
-            )
-            .map(Into::into)
-        }))
+        BoxedFilter(RcType::new(
+            move |env, value, args| -> Result<Value, Error> {
+                f.apply_to(
+                    env,
+                    ArgType::from_value(Some(value))?,
+                    FunctionArgs::from_values(args)?,
+                )
+                .map(Into::into)
+            },
+        ))
     }
 
     /// Applies the filter to a value and argument.
@@ -102,16 +103,20 @@ pub(crate) fn get_builtin_filters() -> BTreeMap<&'static str, BoxedFilter> {
     let mut rv = BTreeMap::new();
     rv.insert("safe", BoxedFilter::new(safe));
     rv.insert("escape", BoxedFilter::new(escape));
+    rv.insert("e", BoxedFilter::new(escape));
     #[cfg(feature = "builtin_filters")]
     {
         rv.insert("lower", BoxedFilter::new(lower));
         rv.insert("upper", BoxedFilter::new(upper));
         rv.insert("replace", BoxedFilter::new(replace));
         rv.insert("length", BoxedFilter::new(length));
+        rv.insert("count", BoxedFilter::new(length));
         rv.insert("dictsort", BoxedFilter::new(dictsort));
         rv.insert("reverse", BoxedFilter::new(reverse));
         rv.insert("trim", BoxedFilter::new(trim));
         rv.insert("join", BoxedFilter::new(join));
+        rv.insert("default", BoxedFilter::new(default));
+        rv.insert("d", BoxedFilter::new(default));
     }
     rv
 }
@@ -123,6 +128,8 @@ pub fn safe(_env: &Environment, v: String) -> Result<Value, Error> {
 }
 
 /// HTML escapes a string.
+///
+/// By default this filter is also registered under the alias `e`.
 pub fn escape(_env: &Environment, v: Value) -> Result<Value, Error> {
     // TODO: this ideally understands which type of escaping is in use
     if v.is_safe() {
@@ -168,6 +175,8 @@ mod builtins {
     }
 
     /// Returns the "length" of the value
+    ///
+    /// By default this filter is also registered under the alias `count`.
     #[cfg_attr(docsrs, doc(cfg(feature = "builtin_filters")))]
     pub fn length(_env: &Environment, v: Value) -> Result<Value, Error> {
         v.len().map(Value::from).ok_or_else(|| {
@@ -254,6 +263,18 @@ mod builtins {
         }
     }
 
+    /// Checks if a string starts with another string.
+    ///
+    /// By default this filter is also registered under the alias `d`.
+    #[cfg_attr(docsrs, doc(cfg(feature = "builtin_filters")))]
+    pub fn default(_: &Environment, value: Value, other: Option<Value>) -> Result<Value, Error> {
+        Ok(if value.is_undefined() {
+            other.unwrap_or_else(|| Value::from(""))
+        } else {
+            value
+        })
+    }
+
     #[test]
     fn test_basics() {
         fn test(_: &Environment, a: u32, b: u32) -> Result<u32, Error> {
