@@ -30,17 +30,16 @@
 //! MiniJinja will perform the necessary conversions automatically via the
 //! [`FunctionArgs`](crate::value::FunctionArgs) trait.
 use std::collections::BTreeMap;
-use std::sync::Arc;
 
 use crate::environment::Environment;
 use crate::error::Error;
-use crate::value::{ArgType, FunctionArgs, Value};
+use crate::value::{ArgType, FunctionArgs, RcType, Value};
 
 type TestFunc =
     dyn Fn(&Environment, Value, Vec<Value>) -> Result<bool, Error> + Sync + Send + 'static;
 
 #[derive(Clone)]
-pub(crate) struct BoxedTest(Arc<TestFunc>);
+pub(crate) struct BoxedTest(RcType<TestFunc>);
 
 /// A utility trait that represents filters.
 pub trait Test<V = Value, Args = Vec<Value>>: Send + Sync + 'static {
@@ -77,13 +76,15 @@ impl BoxedTest {
         V: ArgType,
         Args: FunctionArgs,
     {
-        BoxedTest(Arc::new(move |env, value, args| -> Result<bool, Error> {
-            f.perform(
-                env,
-                ArgType::from_value(Some(value))?,
-                FunctionArgs::from_values(args)?,
-            )
-        }))
+        BoxedTest(RcType::new(
+            move |env, value, args| -> Result<bool, Error> {
+                f.perform(
+                    env,
+                    ArgType::from_value(Some(value))?,
+                    FunctionArgs::from_values(args)?,
+                )
+            },
+        ))
     }
 
     /// Applies the filter to a value and argument.
@@ -110,6 +111,8 @@ pub(crate) fn get_builtin_tests() -> BTreeMap<&'static str, BoxedTest> {
         rv.insert("string", BoxedTest::new(is_string));
         rv.insert("sequence", BoxedTest::new(is_sequence));
         rv.insert("mapping", BoxedTest::new(is_mapping));
+        rv.insert("startingwith", BoxedTest::new(is_startingwith));
+        rv.insert("endingwith", BoxedTest::new(is_endingwith));
     }
     rv
 }
@@ -173,6 +176,18 @@ mod builtins {
         Ok(matches!(v.kind(), ValueKind::Map))
     }
 
+    /// Checks if the value is starting with a string.
+    #[cfg_attr(docsrs, doc(cfg(feature = "builtin_tests")))]
+    pub fn is_startingwith(_env: &Environment, v: String, other: String) -> Result<bool, Error> {
+        Ok(v.starts_with(&other))
+    }
+
+    /// Checks if the value is ending with a string.
+    #[cfg_attr(docsrs, doc(cfg(feature = "builtin_tests")))]
+    pub fn is_endingwith(_env: &Environment, v: String, other: String) -> Result<bool, Error> {
+        Ok(v.ends_with(&other))
+    }
+
     #[test]
     fn test_basics() {
         fn test(_: &Environment, a: u32, b: u32) -> Result<bool, Error> {
